// testing minimal CRC->CR..to..CRC->DR delay
// https://community.st.com/thread/40020-crc-data-ignored-after-crc-reset
// easy to test if you put a write-data-breakpoint to CRC->DR

// original 10.4.2017
// 2nd take, 5.8.2023, https://community.st.com/t5/stm32-mcu-products/erratum-quot-delay-after-an-rcc-peripheral-clock-enabling-quot/td-p/579949

// Includes
#include "stm32f4xx.h"

// --------- app defines

#define NOP_NR 0                // nr of NOPs, 0..2
#define DSB_NR 0                // nr of DSBs, 0..1
#define READ_ENABLE 0           // enable readback of RCC_AxByENR register
#define READ_PERIPHERAL 0       // enable readback of target peripheral register

#define DISABLE_WRITE_BUFFER 0

volatile _Bool    result[8];
volatile uint32_t ticks[8];


// -------- defines


// Joseph Yiu's method
// From http://forums.arm.com/index.php?showtopic=13949
volatile unsigned int *DWT_CYCCNT     = (volatile unsigned int *)0xE0001004; //address of the register
volatile unsigned int *DWT_CONTROL    = (volatile unsigned int *)0xE0001000; //address of the register
volatile unsigned int *SCB_DEMCR         = (volatile unsigned int *)0xE000EDFC; //address of the register


int main(void)
{
    *SCB_DEMCR = *SCB_DEMCR | 0x01000000;
    *DWT_CYCCNT = 0; // reset the counter
    *DWT_CONTROL = *DWT_CONTROL | 1 ; // enable the counter
    // now in gdb, when single-stepping by instructions (si),
    // we can display at each step both the instruction
    //   disp /i $pc
    // and the currect cycle counter value
    //   disp /x *DWT_CYCCNT



#if (0)
  RCC->CFGR |=    RCC_CFGR_HPRE_DIV1    // AHB prescaler set to 1 -> AHB clock = 160MHz

               | RCC_CFGR_PPRE1_DIV4   // APB1 prescaler set to 4 -> APB1 clock = 40MHz (required to be < 42MHz)
               | RCC_CFGR_PPRE2_DIV2;  // APB2 prescaler set to 2 -> APB2 clock = 80MHz (required to be < 84MHz)

    FLASH->ACR = 0
  //    | FLASH_ACR_DCEN
  //    | FLASH_ACR_ICEN
  //    | FLASH_ACR_PRFTEN
  //    | FLASH_ACR_LATENCY_7WS
  //    | FLASH_ACR_LATENCY_2WS
    ;
#else
  RCC->CFGR |=    RCC_CFGR_HPRE_DIV1    // AHB prescaler set to 1 -> AHB clock = 160MHz

               | RCC_CFGR_PPRE1_DIV2   // APB1 prescaler set to 4 -> APB1 clock = 40MHz (required to be < 42MHz)
               | RCC_CFGR_PPRE2_DIV1;  // APB2 prescaler set to 2 -> APB2 clock = 80MHz (required to be < 84MHz)

    FLASH->ACR = 0
      | FLASH_ACR_DCEN
      | FLASH_ACR_ICEN
      | FLASH_ACR_PRFTEN
  //    | FLASH_ACR_LATENCY_7WS
  //    | FLASH_ACR_LATENCY_2WS
    ;
#endif


#if(0) // following is in cq_test1.inc
// RCC_AHB1ENR.GPIOCEN -- RCC offset 0x30, bit 2
  {
    register uint32_t tmp1, tmp2;
    __asm volatile(
      "movs  %[t1], #1    "     "\n\t"
      "ldr   %[t2], [%[p2], #48]    "     "\n\t"
      "orr   %[t2], #4 "     "\n\t"
      "str   %[t2], [%[p2], #48]    "     "\n\t"
#if(NOP_NR >= 1)
      "nop"                     "\n\t"
#endif
#if(NOP_NR >= 2)
      "nop"                     "\n\t"
#endif
#if(DSB_NR)
      "dsb"                     "\n\t"
#endif
#if(READ_ENABLE)
      "ldr   %[t2], [%[p2], #48]    "     "\n\t"  // readback of RCC enable register
#elif(READ_PERIPHERAL)
      "ldr   %[t2], [%[p], #0x14]    "     "\n\t"
#else
    // nothing
#endif
      "str   %[t1], [%[p], #0x14]" "\n\t"

      "nop"                     "\n\t"
      "nop"                     "\n\t"
      "nop"                     "\n\t"

      :[t1] "=&r" (tmp1)
      ,[t2] "=&r" (tmp2)
      :[p]   "r" (GPIOC)
      ,[p2]   "r" (RCC)
    );
  }

  result[test_nr] = (GPIOC->ODR == 1);

  RCC->AHB1RSTR = RCC_AHB1RSTR_GPIOCRST;
  RCC->AHB1RSTR = 0;
  RCC->AHB1ENR &= ~RCC_AHB1ENR_GPIOCEN;
    __asm("nop");
    __asm("nop");
    __asm("nop");
    __asm("nop");
// end of test
#endif


uint32_t test_nr;
register uint32_t tick_start, tick_end;

#if (DISABLE_WRITE_BUFFER)
  SCnSCB->ACTLR = SCnSCB_ACTLR_DISDEFWBUF_Msk;  // reduces nr of required nops
#endif


#undef  NOP_NR
#undef  DSB_NR
#undef  READ_ENABLE
#undef  READ_PERIPHERAL
#define NOP_NR 0                // nr of NOPs, 0..2
#define DSB_NR 0                // nr of DSBs, 0..1
#define READ_ENABLE 0           // enable readback of RCC_AxByENR register
#define READ_PERIPHERAL 0       // enable readback of target peripheral register

test_nr = 0;
#include "cq_test1.inc"

#undef  NOP_NR
#undef  DSB_NR
#undef  READ_ENABLE
#undef  READ_PERIPHERAL
#define NOP_NR 0                // nr of NOPs, 0..2
#define DSB_NR 0                // nr of DSBs, 0..1
#define READ_ENABLE 1           // enable readback of RCC_AxByENR register
#define READ_PERIPHERAL 0       // enable readback of target peripheral register

test_nr = 1;
#include "cq_test1.inc"

#undef  NOP_NR
#undef  DSB_NR
#undef  READ_ENABLE
#undef  READ_PERIPHERAL
#define NOP_NR 0                // nr of NOPs, 0..2
#define DSB_NR 0                // nr of DSBs, 0..1
#define READ_ENABLE 0           // enable readback of RCC_AxByENR register
#define READ_PERIPHERAL 1       // enable readback of target peripheral register

test_nr = 2;
#include "cq_test1.inc"

#undef  NOP_NR
#undef  DSB_NR
#undef  READ_ENABLE
#undef  READ_PERIPHERAL
#define NOP_NR 0                // nr of NOPs, 0..2
#define DSB_NR 0                // nr of DSBs, 0..1
#define READ_ENABLE 0           // enable readback of RCC_AxByENR register
#define READ_PERIPHERAL 0       // enable readback of target peripheral register

test_nr = 3;
#include "cq_test1.inc"


#undef  NOP_NR
#undef  DSB_NR
#undef  READ_ENABLE
#undef  READ_PERIPHERAL
#define NOP_NR 1                // nr of NOPs, 0..2
#define DSB_NR 0                // nr of DSBs, 0..1
#define READ_ENABLE 0           // enable readback of RCC_AxByENR register
#define READ_PERIPHERAL 0       // enable readback of target peripheral register

test_nr = 4;
#include "cq_test1.inc"

#undef  NOP_NR
#undef  DSB_NR
#undef  READ_ENABLE
#undef  READ_PERIPHERAL
#define NOP_NR 2                // nr of NOPs, 0..2
#define DSB_NR 0                // nr of DSBs, 0..1
#define READ_ENABLE 0           // enable readback of RCC_AxByENR register
#define READ_PERIPHERAL 0       // enable readback of target peripheral register

test_nr = 5;
#include "cq_test1.inc"

#undef  NOP_NR
#undef  DSB_NR
#undef  READ_ENABLE
#undef  READ_PERIPHERAL
#define NOP_NR 0                // nr of NOPs, 0..2
#define DSB_NR 1                // nr of DSBs, 0..1
#define READ_ENABLE 0           // enable readback of RCC_AxByENR register
#define READ_PERIPHERAL 0       // enable readback of target peripheral register

test_nr = 6;
#include "cq_test1.inc"

#undef  NOP_NR
#undef  DSB_NR
#undef  READ_ENABLE
#undef  READ_PERIPHERAL
#define NOP_NR 0                // nr of NOPs, 0..2
#define DSB_NR 0                // nr of DSBs, 0..1
#define READ_ENABLE 0           // enable readback of RCC_AxByENR register
#define READ_PERIPHERAL 0       // enable readback of target peripheral register

test_nr = 7;
#include "cq_test1.inc"



__BKPT(0);
/*
  results:
STM32F411RET6  rev.A datecode 414 (2014)
(gdb) p result
$6 = {false, true, true, false, true, true, true, false}
(gdb) p ticks
$7 = {5, 6, 7, 5, 6, 7, 7, 5}
(gdb) x 0xE000ED00
0xe000ed00:     0x410fc241  // Cortex-M4 r0p1

STM32F407VGT6  rev.Z datecode 213 (2012)
(gdb) p result
$1 = {false, false, false, false, false, true, true, false}
(gdb) p ticks
$2 = {6, 7, 7, 6, 6, 8, 9, 6}
(gdb)  x 0xE000ED00
0xe000ed00:     0x410fc241

STM32F429ZIT6U  rev.Y datecode 328 (2013)
(gdb) p result
$1 = {false, false, false, false, false, true, true, false}
(gdb) p ticks
$2 = {6, 7, 7, 6, 6, 8, 9, 6}
(gdb) x 0xE000ED00
0xe000ed00:     0x410fc241

STM32L476VGT6U  rev.4 datecode 543 (2015)
(gdb) p result
$1 = {false, true, false, false, false, true, true, false}
(gdb) p ticks
$2 = {5, 7, 6, 5, 6, 7, 7, 5}
(gdb) x 0xE000ED00
0xe000ed00:     0x410fc241

works?/ticks
  idx  method      'F411   'F407='F427   'L476
0,3,7  none         no/5       no/6       no/5
    1  read RCC    yes/6       no/7      yes/7
    2  read GPIO   yes/7       no/7       no/6
    4  one NOP     yes/6       no/6       no/6
    5  two NOPs    yes/7      yes/8      yes/7
    6  DSB         yes/7      yes/6      yes/7
*/




/*
for GPIO:
  -  two NOPs are enough
     - if write buffer disabled, one nop is enough
  - dsb is enough
  - readback is not enough, even if write buffer disabled
  - one nop before ldr is not enough (it is enough if write buffer disabled), but after ldr is enough
*/



// now experiment with an APB1 peripheral (TIM2):
  {
    register uint32_t tmp1, tmp2;
    __asm volatile(
      "movs  %[t1], #1    "     "\n\t"
      "ldr   %[t2], [%[p2], #64]    "     "\n\t"
      "orr   %[t2], #1 "     "\n\t"
      "str   %[t2], [%[p2], #64]    "     "\n\t"
#if(0)
      "nop"                     "\n\t"
      "nop"                     "\n\t"
#elif(0)
      "dsb"                     "\n\t"
#elif(0)
      "ldr   %[t2], [%[p2], #64]    "     "\n\t"
#elif(0)
      "ldr   %[t2], [%[p], #0]    "     "\n\t"
#else
    // nothing
#endif
      "str   %[t1], [%[p], #0]" "\n\t"

      "nop"                     "\n\t"
      "nop"                     "\n\t"
      "nop"                     "\n\t"

      :[t1] "=&r" (tmp1)
      ,[t2] "=&r" (tmp2)
      :[p]   "r" (TIM2)
      ,[p2]   "r" (RCC)
    );
  }






  {
    register uint32_t tmp1, tmp2;
    __asm volatile(
      "movs  %[t1], #1    "     "\n\t"
      "ldr   %[t2], [%[p2], #48]    "     "\n\t"
      "orr   %[t2], #4096 "     "\n\t"
      "str   %[t2], [%[p2], #48]    "     "\n\t"
#if(0)
      "ldr   %[t2], [%[p2], #48]    "     "\n\t"
#elif(1)
      "dsb"                     "\n\t"
#else
      "nop"                     "\n\t"
      "nop"                     "\n\t"
#endif
      "str   %[t1], [%[p]]" "\n\t"

      "movs  %[t2], #0    "     "\n\t"
      "nop"                     "\n\t"
      "nop"                     "\n\t"
      "str   %[t1], [%[p], #8]" "\n\t"
#if(1)
      "dsb"                     "\n\t"
#else
      "nop"                     "\n\t"
      "nop"                     "\n\t"
#endif
      "str   %[t2], [%[p]]"     "\n\t"
      "nop"                     "\n\t"
      "nop"                     "\n\t"
      "str   %[t1], [%[p], #8]" "\n\t"
      "ldr   %[t1], [%[p], #8]" "\n\t"
      "str   %[t2], [%[p]]"     "\n\t"
      "nop"                     "\n\t"
      "nop"                     "\n\t"
      :[t1] "=&r" (tmp1)
      ,[t2] "=&r" (tmp2)
      :[p]   "r" (CRC)
      ,[p2]   "r" (RCC)
    );
  }

// conclusion:
// - one nop is not enough, two are needed
// - one dsb is enough instead of nops
// - the loadback is not enough, it required the same two nops after it
// - the same applies for the write-RCC_AHB1ENR--to--write-CRC_DR delay, one nop is not enough

// in the following, gcc sometimes reordered the __asm("nop") outside the two volatile writes!

  CRC->CR = 1;
    __asm("nop");
  CRC->DR = 0xFFFFFFFF;
    __asm("nop");




  while(1) {
    __asm("nop");

  }


}






